'use strict';

class nestedSort {

  constructor({
    droppingEdge = 15,
    el
  } = {}) {
    this.selector = el;
    this.sortableList = null;
    this.placeholderUl = null;
    this.placeholderInUse = null;
    this.draggedNode = null;
    this.targetedNode = null;

    this.targetNode = {
      X: null,
      Y: null
    };

    this.distances = {
      droppingEdge,
      droppingEdgeNegative: droppingEdge * -1,
      mouseTo: {
        targetedElTop: undefined
      }
    };

    this.dimensions = {
      targetedEl: {
        H: undefined
      }
    };

    this.cursor = {
      X: null,
      Y: null
    };

    this.classNames = {
      dragged: 'ns-dragged',
      placeholder: 'ns-placeholder',
      targeted: 'ns-targeted',
    };

    this.initDragAndDrop();
  }

  initDragAndDrop() {

    document.addEventListener('dragover', this.dragListener.bind(this), false);

    this.initPlaceholderList();

    this.sortableList = document.getElementById(this.selector);

    this.sortableList.querySelectorAll('li').forEach(el => {
      el.setAttribute('draggable', 'true');

      el.addEventListener('dragstart', this.onDragStart.bind(this), false);
      el.addEventListener('dragenter', this.onDragEnter.bind(this), false);
      el.addEventListener('dragend', this.onDragEnd.bind(this), false);

      this.addListItemStyles(el);
    });
  }

  getComputedStyleValue(el, prop) {
    return window.getComputedStyle(el, null).getPropertyValue(prop);
  }

  addListItemStyles(li) {
    // let's add a move cursor icon if it does not already have a cursor css property
    const cursor = this.getComputedStyleValue(li, 'cursor');
    if (!cursor || cursor === 'auto') {
      li.style.cursor = 'move';
    }
  }

  onDragStart(e) {
    this.draggedNode = e.target;
    this.draggedNode.classList.add(this.classNames.dragged);
    e.dataTransfer.setData('text', 'Drag started!'); // Hack for Firefox!
  }

  onDragOver(e) {
    e.preventDefault(); // prevent default to allow drop
  }

  onDragEnter(e) {
    if (['LI', 'UL'].includes(e.target.nodeName)) {
      e.preventDefault(); // prevent default to allow drop

      if (this.targetedNode) this.targetedNode.classList.remove(this.classNames.targeted);
      this.targetedNode = e.target;
      e.target.classList.add(this.classNames.targeted);

      e.target.addEventListener('dragover', this.onDragOver.bind(this), false);
      e.target.addEventListener('drop', this.onDrop.bind(this), false);
      e.target.addEventListener('dragleave', this.onDragLeave.bind(this), false);
    }
  }

  onDragLeave(e) {
    e.preventDefault();
    e.target.removeEventListener('dragover', this.onDrop);
    e.target.removeEventListener('drop', this.onDrop);
    e.target.removeEventListener('dragleave', this.onDragLeave);
  }

  onDragEnd(e) {
    e.preventDefault();
    this.draggedNode.classList.remove(this.classNames.dragged);
    this.targetedNode.classList.remove(this.classNames.targeted);
    this.cleanupPlaceholderLists();
  }

  onDrop(e) {
    e.preventDefault();
    this.maybeDrop();
    this.cleanupPlaceholderLists();
  }

  dragListener(e) {
    this.updateCoordination(e);
    this.managePlaceholderLists(e);
  }

  updateCoordination(e) {
    this.calcMouseCoords(e);
    this.calcMouseToTargetedElDist();
  }

  maybeDrop(e) {
    if (!this.canBeDropped()) {
      return;
    }

    let dropLocation;
    if (this.targetedNode.nodeName === 'LI' && !this.cursorIsIndentedEnough()) {
      dropLocation = 'before';
    } else if (this.targetedNode.nodeName === 'UL') {
      dropLocation = 'inside';
    }

    if (dropLocation) this.dropTheItem(dropLocation, e);
  }

  dropTheItem(place, e) {
    switch (place) {
      case 'before':
        this.targetedNode.parentNode.insertBefore(this.draggedNode, this.targetedNode);
        break;
      case 'after':
        this.insertAfter(this.draggedNode, this.targetedNode);
        break;
      case 'inside':
        this.targetedNode.appendChild(this.draggedNode);
        break;
    }
  }

  calcMouseCoords(e) {
    // we're having the client coords because on the next lines, we use getBoundingClientRect which behaves in the same way
    this.cursor.X = e.clientX;
    this.cursor.Y = e.clientY;
  }

  calcMouseToTargetedElDist() {
    if (!this.targetedNode) {
      return;
    }

    let offset = this.targetedNode.getBoundingClientRect();
    this.targetNode.X = offset.left;
    this.targetNode.Y = offset.top;

    let result = this.targetNode.Y - this.cursor.Y;
    this.distances.mouseTo.targetedElTop = result;
    this.distances.mouseTo.targetedElTopAbs = Math.abs(result);
    this.dimensions.targetedEl.H = this.targetedNode.clientHeight;
    this.distances.mouseTo.targetedElBot = this.distances.mouseTo.targetedElTopAbs - this.dimensions.targetedEl.H;
  }

  areNested(child, parent) {
    let isChild = false;
    parent.querySelectorAll('li').forEach(li => {
      if (li === child) {
        isChild = true;
      }
    });
    return isChild;
  }

  cursorIsIndentedEnough() {
    return this.cursor.X - this.targetNode.X > 50;
  }

  mouseIsTooCloseToTop() {
    return this.cursor.Y - this.targetNode.Y < this.distances.droppingEdge;
  }

  insertAfter(newNode, referenceNode) {
    referenceNode.parentNode.insertBefore(newNode, referenceNode.nextSibling);
  }

  managePlaceholderLists(e) {

    let actions = this.analysePlaceHolderSituation(e);

    actions.forEach(action => {
      switch (action) {
        case 'add':
          this.cleanupPlaceholderLists();
          this.addPlaceholderList();
          break;
        case 'cleanup':
          this.cleanupPlaceholderLists();
          break;
      }
    });
  }

  targetedNodeIsPlaceholder() {
    return this.targetedNode.nodeName === 'UL' && this.targetedNode.classList.contains(this.classNames.placeholder);
  }

  analysePlaceHolderSituation(e) {
    if (!this.targetedNode || this.areNested(this.targetedNode, this.draggedNode)) {
      return [];
    }

    let actions = [];

    if (!this.cursorIsIndentedEnough() || this.mouseIsTooCloseToTop()) {
      if (!this.targetedNodeIsPlaceholder()) {
        actions.push('cleanup');
      }
    } else if (this.targetedNode !== this.draggedNode
      && this.targetedNode.nodeName === 'LI'
      && !this.targetedNode.querySelectorAll('ul').length) {
      actions.push('add');
    }

    return actions;
  }

  addPlaceholderList() {
    const list = this.getPlaceholderList();
    list.style.minHeight = '0';
    this.targetedNode.appendChild(list);
    list.style.transition = 'min-height ease .2s';
    list.style.minHeight = `${this.draggedNode.offsetHeight}px`;
  }

  targetNodeIsIdentified() {
    return !!this.targetedNode;
  }

  canBeDropped() {
    let result = true;

    result &= this.targetNodeIsIdentified() && this.targetedNode !== this.draggedNode;
    result &= this.targetNodeIsIdentified() && !(this.targetedNode.nodeName === 'UL' && this.targetedNode.querySelectorAll('li').length);
    result &= !this.areNested(this.targetedNode, this.draggedNode);

    return result;
  }

  cleanupPlaceholderLists() {
    this.sortableList.querySelectorAll('ul').forEach(ul => {
      if (!ul.querySelectorAll('li').length) {
        ul.remove();
      } else if (ul.classList.contains(this.classNames.placeholder)) {
        ul.classList.remove(this.classNames.placeholder);
      }
    });
  }

  initPlaceholderList() {
    this.placeholderUl = document.createElement('ul');
    this.placeholderUl.classList.add(this.classNames.placeholder);
  }

  getPlaceholderList() {
    this.placeholderInUse = this.placeholderUl.cloneNode(true);
    return this.placeholderInUse;
  }
}

module.exports = nestedSort;
